import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';

typedef DraggableItemBuilder<T> = Widget Function(
  BuildContext context,
  T item,
  int index,
);

class DraggableGridView<T> extends StatefulWidget {
  final List<T>? initialValue;
  final DraggableItemBuilder<T> builder;
  final Duration animateDuration;
  final GridController<T>? controller;
  final SliverGridDelegate gridDelegate;
  final DraggableConfiguration<T> draggableConfiguration;
  final ValueChanged<List<T>>? onAccept;
  final void Function(List<T>? items, int index)? onDragCompleted;
  final Curve? slideCurve;

  const DraggableGridView({
    this.draggableConfiguration = const DraggableConfiguration.disabled(),
    required this.builder,
    this.initialValue,
    this.controller,
    required this.gridDelegate,
    this.animateDuration = const Duration(milliseconds: 100),
    this.onAccept,
    this.onDragCompleted,
    Key? key,
    this.slideCurve,
  }) : super(key: key);
  @override
  State<StatefulWidget> createState() => _DraggableGridViewState<T>();
}

class _DraggableGridViewState<T> extends State<DraggableGridView<T>> {
  GridController<T>? _controller;
  GridController<T> get _effectiveController =>
      widget.controller ?? _controller!;

  @override
  void initState() {
    super.initState();
    if (widget.controller == null) {
      _createLocalController(widget.initialValue ?? []);
    }
    _effectiveController.addListener(onValueChanged);
  }

  void onValueChanged() {
    if (mounted) {
      setState(() {});
    }
  }

  @override
  void dispose() {
    widget.controller?.removeListener(onValueChanged);
    _controller?.dispose();
    super.dispose();
  }

  void _createLocalController(List<T> items) {
    _controller = GridController<T>(items);
  }

  @override
  void didUpdateWidget(DraggableGridView<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.controller != oldWidget.controller) {
      if (oldWidget.controller != null && widget.controller == null) {
        _createLocalController(oldWidget.controller!.value);
      }

      if (widget.controller != null) {
        if (oldWidget.controller == null) {
          _controller!.dispose();
          _controller = null;
        }
      }
    }
  }

  @override
  Widget build(BuildContext context) {
    return OrientationBuilder(builder: (context, o) {
      return GridView.builder(
        shrinkWrap: true,
        gridDelegate: widget.gridDelegate,
        itemCount: _effectiveController.value.length,
        itemBuilder: (context, index) {
          return DraggableGridItem<T>(
            widget.animateDuration,
            _effectiveController.createIfNotExists(index),
            _effectiveController,
            widget.builder,
            _effectiveController.value[index],
            widget.draggableConfiguration,
            widget.onAccept,
            widget.onDragCompleted,
            widget.slideCurve,
          );
        },
      );
    });
  }
}

class GridController<T> extends ValueNotifier<List<T>> {
  final Map<int, _DraggableController> _controllers = {};
  final Map<int, VoidCallback> _completeCallback = {};
  List<T>? _animateItems;

  int _gen = 0;

  GridController(List<T> items) : super(items);

  _DraggableController createIfNotExists(int index) {
    return _controllers.putIfAbsent(index, () => _DraggableController(index));
  }

  bool canAccept(dynamic data) {
    if (data == null) {
      return false;
    }
    return _controllers.values.contains(data);
  }

  bool get _accepted => _controllers.values.any((element) => element.accepted);

  @override
  void dispose() {
    _clearAnimations();
    super.dispose();
  }

  void _clearAnimations() {
    _completeCallback.clear();
    for (final _DraggableController element in _controllers.values) {
      element.dispose();
    }
    _controllers.clear();
  }

  void _drag(int start, int end) {
    _dragAnimate(start, end);
  }

  void removeAnimate(List<int> indexes, ValueChanged<List<T>> onCompleted) {
    final List<T>? items = _getRemovedList(indexes);
    if (items == null) {
      return;
    }
    final int gen = ++_gen;
    for (final element in _controllers.values) {
      final T _item = value[element.index];
      final int index = items.indexOf(_item);
      element._currentIndex = index;
    }
    _animate(gen, () {
      onCompleted(items);
    });
  }

  @override
  set value(List<T> newValue) {
    _clearAnimations();
    _animateItems = null;
    super.value = newValue;
  }

  List<T>? _getRemovedList(List<int> indexes) {
    final List<int> validIndexes = indexes
        .where((element) => element >= 0 && element < value.length)
        .toSet()
        .toList()
      ..sort((a, b) => -a.compareTo(b));
    if (validIndexes.isEmpty) {
      return null;
    }
    _animateItems ??= List.of(value);
    validIndexes
        .map((e) => _controllers[e]!.currentIndex)
        .where((element) => element != -1)
        .forEach(_animateItems!.removeAt);
    return _animateItems;
  }

  /// [start] the index of DraggableItem
  ///
  /// [end] the **current index** of DraggableItem
  void _dragAnimate(int start, int end) {
    final int gen = ++_gen;
    _animateItems ??= List.of(value);
    final T item = value[start];
    _animateItems!
        .insert(end, _animateItems!.removeAt(_animateItems!.indexOf(item)));
    for (final element in _controllers.values) {
      final T _item = value[element.index];
      element._currentIndex = _animateItems!.indexOf(_item);
    }
    _animate(gen);
  }

  void _animate(int gen, [VoidCallback? onComplete]) {
    final List<int> animationIndexs = [];
    for (final _DraggableController controller in _controllers.values) {
      if (controller.currentIndex == -1) {
        controller.value = null;
      } else if (controller.index != controller.currentIndex) {
        final Offset targetPosition =
            _getPixelPosition(controller.currentIndex);
        final Offset position = _getPixelPosition(controller.index);
        animationIndexs.add(controller.index);
        controller.value = _AnimateInfo(
            Offset(
              (targetPosition.dx - position.dx) / controller.size.width,
              (targetPosition.dy - position.dy) / controller.size.height,
            ),
            gen);
      } else {
        if (controller.value?.offset != Offset.zero) {
          animationIndexs.add(controller.index);
          controller.value = _AnimateInfo(Offset.zero, gen);
        }
      }
    }
    if (animationIndexs.isNotEmpty) {
      _completeCallback[gen] = () {
        onComplete?.call();
      };
    } else {
      if (onComplete != null) {
        onComplete.call();
      }
    }
  }

  void _onAnimateEnd(int gen, int index) {
    _completeCallback.remove(gen)?.call();
  }

  Offset _getPixelPosition(int index) {
    for (final _DraggableController controller in _controllers.values) {
      if (controller.index == index) {
        final RenderObject? renderObject =
            controller.context.findRenderObject();
        if (renderObject != null) {
          return (renderObject as RenderBox).localToGlobal(Offset.zero);
        }
      }
    }
    throw Exception('can not get pixel position at $index');
  }
}

class DraggableGridItem<T> extends StatefulWidget {
  final _DraggableController controller;
  final GridController<T> gridController;
  final DraggableItemBuilder<T> builder;
  final Duration duration;
  final T item;
  final DraggableConfiguration<T> draggableConfiguration;

  /// called when drag canceled or accepted
  final ValueChanged<List<T>>? onAccept;

  /// called when draggable item is accepted by **OTHER** [DragTarget]
  final void Function(List<T>? items, int index)? onDragCompleted;
  final Curve? slideCurve;
  DraggableGridItem(
    this.duration,
    this.controller,
    this.gridController,
    this.builder,
    this.item,
    this.draggableConfiguration,
    this.onAccept,
    this.onDragCompleted,
    this.slideCurve,
  ) : super(key: controller.key);

  @override
  State<StatefulWidget> createState() => _DraggableItemState<T>();
}

class _DraggableItemState<T> extends State<DraggableGridItem<T>>
    with SingleTickerProviderStateMixin {
  final ValueNotifier<Size?> _sizeNotifier = _QuietlySizeNotifier(null);
  late _AnimateInfo? _animateInfo = widget.controller.value!;
  @override
  void initState() {
    super.initState();
    widget.controller.addListener(() {
      if (mounted) {
        setState(() {
          _animateInfo = widget.controller.value;
        });
      }
    });
  }

  void onAnimateEnd() {
    widget.gridController
        ._onAnimateEnd(_animateInfo!.gen, widget.controller.index);
  }

  @override
  void dispose() {
    _sizeNotifier.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    WidgetsBinding.instance!.addPostFrameCallback((t) {
      if (!mounted) {
        return;
      }
      final RenderBox referenceBox = context.findRenderObject()! as RenderBox;
      widget.controller.context = context;
      widget.controller.size = referenceBox.size;
      _sizeNotifier.value = referenceBox.size;
    });
    final Widget child =
        widget.builder(context, widget.item, widget.controller.index);

    final Widget transition = _animateInfo != null
        ? AnimatedSlide(
            offset: _animateInfo!.offset,
            onEnd: onAnimateEnd,
            duration: widget.duration,
            child: child,
            curve: widget.slideCurve ?? Curves.linear,
          )
        : const SizedBox.shrink();

    if (!widget.draggableConfiguration
        .enableChecker(widget.item, widget.controller.index)) {
      return transition;
    }

    void _onAccept() {
      if (widget.gridController._animateItems != null) {
        widget.onAccept?.call(widget.gridController._animateItems!);
      }
    }

    void _onDragComplete(int index) {
      final bool accepted = widget.gridController._accepted;
      if (!accepted) {
        widget.onDragCompleted
            ?.call(widget.gridController._animateItems, index);
      } else {
        _onAccept();
      }
    }

    return DragTarget<_DraggableController>(
      onWillAccept: (data) {
        if (!widget.gridController.canAccept(data)) {
          return false;
        }
        widget.gridController._drag(data!.index, widget.controller.index);
        return true;
      },
      onAccept: (data) {
        widget.controller.accepted = true;
      },
      builder: (context, candidateData, rejectedData) {
        return LongPressDraggable(
          maxSimultaneousDrags: 1,
          delay: widget.draggableConfiguration.longPressDelay,
          childWhenDragging:
              widget.draggableConfiguration.childWhenDraggingBuilder?.call(
                      context, widget.item, widget.controller.index, child) ??
                  const SizedBox.shrink(),
          feedback: ValueListenableBuilder<Size?>(
            builder: (context, size, _child) {
              return Material(
                color: Colors.transparent,
                child: widget.draggableConfiguration.feedbackBuilder?.call(
                        context,
                        widget.item,
                        widget.controller.index,
                        child,
                        size) ??
                    SizedBox(
                      width: size?.width,
                      height: size?.height,
                      child: child,
                    ),
              );
            },
            valueListenable: _sizeNotifier,
          ),
          data: widget.controller,
          onDragStarted: () {
            widget.draggableConfiguration.onDragStart
                ?.call(widget.controller.index);
          },
          onDragCompleted: () {
            if (!mounted) {
              return;
            }
            _onDragComplete(widget.controller.index);
          },
          onDraggableCanceled: (velocity, offset) {
            if (!mounted) {
              return;
            }
            _onAccept();
          },
          onDragEnd: (d) {
            if (!mounted) {
              return;
            }
            widget.draggableConfiguration.onDragEnd
                ?.call(widget.controller.index);
          },
          child: transition,
        );
      },
    );
  }
}

class _DraggableController extends ValueNotifier<_AnimateInfo?> {
  /// grid item index
  final int index;

  final Key key = UniqueKey();

  int? _currentIndex;

  /// current index after drag
  int get currentIndex => _currentIndex ?? index;

  bool accepted = false;

  late BuildContext context;
  late Size size;

  _DraggableController(
    this.index,
  ) : super(const _AnimateInfo(Offset.zero, 0));
}

@immutable
class _AnimateInfo {
  final Offset offset;
  final int gen;

  const _AnimateInfo(this.offset, this.gen);

  @override
  bool operator ==(Object o) =>
      o is _AnimateInfo && gen == o.gen && offset == o.offset;
  @override
  int get hashCode => Object.hash(offset, gen);
}

class _QuietlySizeNotifier extends ValueNotifier<Size?> {
  _QuietlySizeNotifier(Size? value) : super(value);
  bool _disposed = false;
  @override
  void dispose() {
    _disposed = true;
    super.dispose();
  }

  @override
  void removeListener(VoidCallback listener) {
    if (!_disposed) {
      super.removeListener(listener);
    }
  }

  @override
  void addListener(VoidCallback listener) {
    if (!_disposed) {
      super.addListener(listener);
    }
  }

  @override
  void notifyListeners() {
    if (!_disposed) {
      super.notifyListeners();
    }
  }
}

class DraggableConfiguration<T> {
  final bool Function(T item, int index) enableChecker;
  final Duration longPressDelay;
  final Widget Function(BuildContext context, T item, int index, Widget child)?
      childWhenDraggingBuilder;
  final Widget Function(
          BuildContext context, T item, int index, Widget child, Size? size)?
      feedbackBuilder;
  final ValueChanged<int>? onDragStart;
  final ValueChanged<int>? onDragEnd;

  static bool _enableAll(dynamic item, int index) {
    return true;
  }

  static bool _disableAll(dynamic item, int index) {
    return false;
  }

  DraggableConfiguration({
    this.enableChecker = _enableAll,
    this.childWhenDraggingBuilder,
    this.feedbackBuilder,
    this.longPressDelay = kLongPressTimeout,
    this.onDragStart,
    this.onDragEnd,
  });

  const DraggableConfiguration.disabled()
      : enableChecker = _disableAll,
        longPressDelay = kLongPressTimeout,
        childWhenDraggingBuilder = null,
        feedbackBuilder = null,
        onDragStart = null,
        onDragEnd = null;
}
